
/*
 * GENCAT_PPMX1 - Generate PPMX catalog (first phase)
 * Made by Daniel E. Severin (CONICET) in 2015-2017
 */

#include "common.h"
#include <unordered_map>

#define MAXPPMXSTAR 150000

struct PPMXstar_struct PPMXstar[MAXPPMXSTAR];
int PPMXstars; /* number of PPMX stars */

/*
 * main
 */
int main(int argc, char** argv)
{
	FILE *stream, *stream_csv, *stream_ini;
	char buffer[1024];
	char cell[256];
	std::unordered_map<std::string, int> namePPMX; // Hash table of PPMX stars

	printf("GENCAT_PPMX1 - Generate PPMX catalog (first phase).\n");
    printf("Made by Daniel E. Severin (CONICET) in 2015-2017.\n");

    /* read Xmatch cross-table */
	stream = fopen("tables\\xmatch_result.dat", "rt");
	if (stream == NULL) {
		perror("Error reading Xmatch file");
		exit(1);
	}

	/* at the same time, write a CSV table for being cross-identified with APASS-DR9 */
	stream_csv = fopen("tables\\ppmx_xmatch.csv", "wt");
	if (stream_csv == NULL) {
		perror("Error creating csv file");
		exit(1);
	}
	fprintf(stream_csv, "RAJ2000,DEJ2000,id\n");

	/* and the initial cross-identification (pairs id_CD, id_PPMX) */
	stream_ini = fopen("tables\\initial_cross.txt", "wt");
	if (stream_ini == NULL) {
		perror("Error creating initial cross-ident. file");
		exit(1);
	}

	PPMXstars = 0;
	fgets(buffer, 1023, stream); /* discard the first line */

	while (fgets(buffer, 1023, stream) != NULL) {
		/* read CD ident. number */
		readfield(buffer, cell, 38, 5);
		int cd_id = atoi(cell);

		/* read PPMX name */
		readfield(buffer, cell, 44, 15);
		std::unordered_map<std::string, int>::const_iterator got = namePPMX.find(cell);
		int ppmx_id = -1;
		if (got != namePPMX.end()) {
			/* use the original id */
			ppmx_id = got->second;
		}
		else {
			/* it is a new star, insert it into the table and store data */
			if (PPMXstars == MAXPPMXSTAR) bye("Maximum number of PPMX stars reached!\n");
			strcpy(PPMXstar[PPMXstars].ppmx_name, cell);
			ppmx_id = PPMXstars;
			namePPMX.insert({ cell, ppmx_id });

			/* read R.A. and Decl. for J2000.0 */
			readfield(buffer, cell, 60, 10);
			float RA2000 = atof(cell);
			readfield(buffer, cell, 71, 10);
			float DE2000 = atof(cell);

			/* read C magnitude */
			readfield(buffer, cell, 226, 6);
			if (cell[0] == ' ') {
				printf("No Cmag found for PPMX%s :(", PPMXstar[PPMXstars].ppmx_name);
				bye("exiting!\n");
			}
			float cmag = atof(cell);

			/* read Johnson V magnitude */
			readfield(buffer, cell, 121, 6);
			float vmag = 15.0;
			bool Vmag_assigned = true;
			if (cell[0] == ' ') {
				/* if magnitude is absent, write the register in CSV file for further cross-ident. with APASS */
				fprintf(stream_csv, "%f,%f,%d\n", RA2000, DE2000, ppmx_id);
				Vmag_assigned = false;
			}
			else vmag = atof(cell);

			/* read proper motion in mas/yr */
			readfield(buffer, cell, 184, 7);
			double pm_ra = atof(cell);
			pm_ra /= (1000.0 * 3600.0) * dcos(DE2000); /* convert from mas*cos(decl)/yr to deg/yr */
			readfield(buffer, cell, 192, 7);
			double pm_decl = atof(cell);
			pm_decl /= 1000.0 * 3600.0; /* convert from mas/yr to deg/yr */

			/* convert coord. to 1875.0: since we need besselian epoch in J2000 format, we use SOFA formula:
			* B = 1900 + (2451545 - 2415020.31352) / 365.242198781 = 2000.001278 */
			double RA1875 = RA2000;
			double DE1875 = DE2000;
			wcsconp(WCS_J2000, WCS_B1950, 0.0, 1875.0, 2000.001278, 1875.0, &RA1875, &DE1875, &pm_ra, &pm_decl);

			/* save more data in PPMX structure */
			PPMXstar[PPMXstars].RA2000 = RA2000;
			PPMXstar[PPMXstars].DE2000 = DE2000;
			PPMXstar[PPMXstars].RA1875 = RA1875;
			PPMXstar[PPMXstars].DE1875 = DE1875;
			PPMXstar[PPMXstars].cmag = cmag;
			PPMXstar[PPMXstars].vmag = vmag;
			PPMXstar[PPMXstars].mag = 0.0; /* <--- to be filled in 2nd. phase */
			PPMXstar[PPMXstars].Vmag_assigned = Vmag_assigned;
			PPMXstars++;
		}

		/* write the pair in the initial cross-ident. file */
		fprintf(stream_ini, "%d,%d\n", cd_id, ppmx_id);
	}
	printf("Number of PPMX stars read: %d\n", PPMXstars);

	fclose(stream_ini);
	fclose(stream_csv);
	fclose(stream);

	/* also save PPMX catalog in binary format */
	stream = fopen("ppmx_data.bin", "wb");
	if (stream == NULL) {
		perror("Error creating bin file");
		exit(1);
	}
	fwrite(&PPMXstars, sizeof(int), 1, stream);
	fwrite(PPMXstar, sizeof(struct PPMXstar_struct), PPMXstars, stream);
	fclose(stream);
	return 0;
}
